;
;  iodbc.def
;
;  $Id: iodbc.def,v 1.1.1.1 2002/04/08 22:48:10 miner Exp $
;
;  DLL link script
;
;  The iODBC driver manager.
;  
;  Copyright (C) 1995 by Ke Jin <kejin@empress.com> 
;  Copyright (C) 1996-2002 by OpenLink Software <iodbc@openlinksw.com>
;  All Rights Reserved.
;
;  This software is released under the terms of either of the following
;  licenses:
;
;      - GNU Library General Public License (see LICENSE.LGPL) 
;      - The BSD License (see LICENSE.BSD).
;
;  While not mandated by the BSD license, any patches you make to the
;  iODBC source code may be contributed back into the iODBC project
;  at your discretion. Contributions will benefit the Open Source and
;  Data Access community as a whole. Submissions may be made at:
;
;      http://www.iodbc.org
;
;
;  GNU Library Generic Public License Version 2
;  ============================================
;  This library is free software; you can redistribute it and/or
;  modify it under the terms of the GNU Library General Public
;  License as published by the Free Software Foundation; either
;  version 2 of the License, or (at your option) any later version.
;
;  This library is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;  Library General Public License for more details.
;
;  You should have received a copy of the GNU Library General Public
;  License along with this library; if not, write to the Free
;  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
;
;  The BSD License
;  ===============
;  Redistribution and use in source and binary forms, with or without
;  modification, are permitted provided that the following conditions
;  are met:
;
;  1. Redistributions of source code must retain the above copyright
;     notice, this list of conditions and the following disclaimer.
;  2. Redistributions in binary form must reproduce the above copyright
;     notice, this list of conditions and the following disclaimer in
;     the documentation and/or other materials provided with the
;     distribution.
;  3. Neither the name of OpenLink Software Inc. nor the names of its
;     contributors may be used to endorse or promote products derived
;     from this software without specific prior written permission.
;
;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
;  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
;  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;

LIBRARY		IODBC.DLL
DESCRIPTION	'iODBC Driver Manager V2.50 (LGPL)'

CODE		MOVEABLE DISCARDABLE
DATA		MOVEABLE MULTIPLE

HEAPSIZE	10000

EXPORTS
    SQLAllocConnect	@1
    SQLAllocEnv		@2
    SQLAllocStmt	@3
    SQLBindCol		@4
    SQLCancel		@5
    SQLColAttributes	@6
    SQLConnect		@7
    SQLDescribeCol	@8
    SQLDisconnect	@9
    SQLError		@10
    SQLExecDirect	@11
    SQLExecute		@12
    SQLFetch		@13
    SQLFreeConnect	@14
    SQLFreeEnv		@15
    SQLFreeStmt		@16
    SQLGetCursorName	@17
    SQLNumResultCols	@18
    SQLPrepare		@19
    SQLRowCount		@20
    SQLSetCursorName	@21
    SQLSetParam		@22
    SQLTransact		@23

    SQLColumns		@40
    SQLDriverConnect	@41
    SQLGetConnectOption	@42
    SQLGetData		@43
    SQLGetFunctions	@44
    SQLGetInfo		@45
    SQLGetStmtOption	@46
    SQLGetTypeInfo	@47
    SQLParamData	@48
    SQLPutData		@49
    SQLSetConnectOption	@50
    SQLSetStmtOption	@51
    SQLSpecialColumns	@52
    SQLStatistics	@53
    SQLTables		@54

    SQLBrowseConnect	@55
    SQLColumnPrivileges	@56
    SQLDataSources	@57
    SQLDescribeParam	@58
    SQLExtendedFetch	@59
    SQLForeignKeys	@60
    SQLMoreResults	@61
    SQLNativeSql	@62
    SQLNumParams	@63
    SQLParamOptions	@64
    SQLPrimaryKeys	@65
    SQLProcedureColumns	@66
    SQLProcedures	@67
    SQLSetPos		@68
    SQLSetScrollOptions	@69
    SQLTablePrivileges	@70
    SQLDrivers		@71
    SQLBindParameter	@72

    LoadByOrdinal	@199
